/****************************************************************************
** When compiling using DJGPP:                                             **
**    GCC VOXELS.C -O VOXELS.EXE                                           **
*****************************************************************************
** When compiling using Turbo/Borland C:                                   **
**  use at least a COMPACT memory model to run this.                       **
*****************************************************************************
** Demonstration of displaying a voxel heightscape                         **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/

#include <conio.h>
#include <stdio.h>
#include <dos.h>
#include <malloc.h>
#include <math.h>
#include <mem.h>

#ifdef __DJGPP
  #include <stdlib.h>
  #include <string.h>
  #include <sys\movedata.h>
  #include <sys\nearptr.h>
  #include <pc.h>
  #include <go32.h>
  #include <dpmi.h>
#endif

  /*This controls the maximum distance displayed. If you increase this,
  you will see more of the landscape, but it will also be slower*/
  const Depth = 0x7F;

  #ifdef __DJGPP
    //Height map
    unsigned char               HMap [65536];
    //Texture map
    unsigned char               TMap [65536];
    //The buffer
    unsigned char             Buffer [64000];
  #else
    //Height map
    unsigned char far                  *HMap;
    //Texture map
    unsigned char far                  *TMap;
    //The buffer
    unsigned char far                *Buffer;
  #endif
  //The screen itself
  unsigned char                      *Screen;
  //Cosine and Sine tables
  int               CosT [2048], SinT [2048];
  //Distance compensation table
  int                           DComp [0x81];
  //Camera information
  unsigned int                   x, y, Angle;
  int                                 Height;
  //Keyboard input
  char                                   Key;

//This sets the display to VGA 320x200 in 256 colours
void VGAScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x13;
  int86(0x10, &r, &r);
}

//This resets the display to text mode
void TextScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x3;
  int86(0x10, &r, &r);
}

//This sets a DAC register to a specific Red Green Blue-value
void SetDAC (unsigned char DAC, unsigned char R, unsigned char G, unsigned char B)
{
  outportb (0x3C8, DAC);
  outportb (0x3C9, R);
  outportb (0x3C9, G);
  outportb (0x3C9, B);
}

//Draws a vertical line down the screen
void DrawVLine (int x1, int y1, int y2, char c)
{
  unsigned int o;
  //o = y1 * 320 + x;
  o = (y1 << 8) + (y1 << 6) + x1;
  //For every pixel along the line
  for (x1 = y1; x1 <= y2; x1++) {
    //Set pixel and move down one
    Buffer [o] = c;
    o += 320;
  }
}

//Casts a ray from specified position and renders result
void Ray (unsigned int a, unsigned int x, unsigned int y, unsigned int sX)
{
  long  deltax, deltay, p, o, d;
  int                      minY;
  long            tx, ty, h, y1;

  deltax = CosT [a];
  deltay = SinT [a];
  minY = 199;
  d = 0;
  tx = x;
  ty = y;

  do {
    tx += deltax;
    ty += deltay;
    //New distance
    d++;
    //Calculate offset into map
    o = (ty & 0xFF00) | ((tx & 0xFF00) >> 8);
    h = HMap [o] - Height;
    //Calculate height
    y1 = DComp [d - 1] - (h << 5) / d;
    if (y1 < minY) {
      //Draw voxel
      DrawVLine (sX, y1, minY - 1, TMap [o]);
      minY = y1;
    }
  } while (d < Depth);
}

void DrawView ()
{
  int a, i;

  //Fill every column on-screen
  for (i = 0; i < 320; i++) {
    //Calculate ray angle depending on view angle
    a = (Angle + i + 1888) & 2047;
    //Cast the ray
    Ray (a, x, y, i);
  }
}

void InitMap ()
{
  int DAC;
  unsigned char Red, Green, Blue, Dummy [4];
  FILE *f;

  #ifndef __DJGPP
  //Assign 64K for a 256x256 texture map
  HMap = (unsigned char far *)farmalloc (65536);

  //Assign 64K for a 256x256 height map
  TMap = (unsigned char far *)farmalloc (65536);
  #endif

  //Load in the file "GROUND.BMP"
  f = fopen ("GROUND.BMP", "rb");
  //The palette starts at byte 54
  fseek (f, 54, SEEK_SET);

  //Read palette
  for (DAC = 0; DAC <= 255; DAC++) {
    //.BMP format saves 4 bytes per colour
    fread (Dummy, 1, 4, f);
    //In reverse order, and in the range of 0..255 instead of 0..63
    Blue  = Dummy [0] >> 2;
    Green = Dummy [1] >> 2;
    Red   = Dummy [2] >> 2;
    //Set the palette
    SetDAC (DAC, Red, Green, Blue);
  }

  //Load in the texture map
  #ifdef __DJGPP
    fread (TMap, 1, 65535, f);
  #else
    fread (&TMap[0], 1, 65535, f);
  #endif
  //And close the file
  fclose (f);

  //Load in the file "HEIGHT.BMP"
  f = fopen ("HEIGHT.BMP", "rb");
  //Skip the palette for this one
  fseek (f, 1078, SEEK_SET);
  #ifdef __DJGPP
    fread (HMap, 1, 65535, f);
  #else
    fread (&HMap[0], 1, 65535, f);
  #endif
  fclose (f);
}

//This void calculates some lookup table
void InitTables ()
{
  unsigned int a;
  float result;

  for (a = 0; a < 2048; a++) {
    //Precalculate cosine
    result = cos ((float)a * M_PI / 1024.0) * 256.0;
    CosT [a] = (int)result;
    //and sine
    result = sin ((float)a * M_PI / 1024.0) * 256.0;
    SinT [a] = (int)result;
  }
  //Precalculate distance compensation table
  for (a = 1; a <= Depth + 1; a++) { DComp [a - 1] = (1000 / (long)a) + 100; }
}

void main ()
{
  //Precalculate sine and cosine
  InitTables ();
  //Set up player
  x = 0x8000;
  y = 0x8000;
  Angle = 600;
  //Switch to graphics mode
  VGAScreen ();
  #ifdef __DJGPP
    __djgpp_nearptr_enable ();
    Screen = (char *) (__djgpp_conventional_base + 0xA0000);
  #else
    Screen = (unsigned char *)MK_FP (0xA000, 0000);
    Buffer = (unsigned char far *)farmalloc (66000);
  #endif

  //Create the map
  InitMap ();

  do {

    //Adjust camera height according to height map
    Height = HMap [(y & 0xFF00) | ((x & 0xFF00) >> 8)];

    //Clear the buffer
    memset (Buffer, 0, 64000);
    //Draw a screen
    DrawView ();
    //Copy the buffer to the screen
    memcpy (Screen, Buffer, 64000);

    //Get keyboard input
    Key = getch ();

    if (Key == 0) switch (getch ()) {
      case 72: //move forward
             x += CosT [Angle];
             y += SinT [Angle];
             break;
      case 75: //turn left
             Angle = (Angle + 2032) & 2047;
             break;
      case 77: //turn right
             Angle = (Angle + 16) & 2047;
             break;
      case 80: //move backward
             x -= CosT [Angle];
             y -= SinT [Angle];
             break;
    }

  } while (Key != 27);
  //Switch back to text
  TextScreen ();
}